/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ABSTRACTTREEITEM_H
#define ABSTRACTTREEITEM_H

#include <QtCore/QStringList>
#include <QtCore/QDebug>
#include <QtCore/QVariant>

#include <map>

namespace NSCore {

	/**
	 * @brief Abstract class for TreeItems that are no QObjects
	 *
	 * This provides the basics for named trees with object that should not be
	 * QObjects. For example results and other values that do not need signals
	 * or slots.
	 */
	class AbstractTreeItem
	{
		private:
			/**
			 * @brief Disabled copy-constructor
			 *
			 * Copying an item doesn't make sense without a new parent...
			 */
			AbstractTreeItem( AbstractTreeItem& ) {}

			/**
			 * @brief type for the list of childs.
			 *
			 * For convinience
			 */
			typedef std::map<QString,AbstractTreeItem*> childlist_t;

		protected:
			/**
			 * @brief Construct a new item as child of a given parent
			 *
			 * If the parent is 0, this will/can be the root of a new tree.
			 *
			 * @param name The name of the new item
			 * @param parent The parent item for this new item
			 */
			AbstractTreeItem( QString name, AbstractTreeItem* parent );
			/**
			 * @brief Construct a new item by copying from another item
			 *
			 * As the new item needs another parent then the original item, this
			 * new parent has to be provided.
			 *
			 * @param item The item that is to be copied
			 * @param parent The parent for this new item
			 */
			AbstractTreeItem( AbstractTreeItem& item, AbstractTreeItem* parent );

			/**
			 * @brief Destructor
			 *
			 * Removes itself from the parent ( if there is a parent ) and deletes all the children
			 */
			virtual ~AbstractTreeItem();

		public:

			/**
			 * @brief Return the number of childs of this item
			 */
			virtual int childs() const { return _childs.size(); }

			/**
			 * @brief Return the name of this item
			 */
			virtual QString name() const { return _name; }

			/**
			 * @brief Return a list of names of the children
			 *
			 * This returns only the names of the direct children, no traversing
			 * on the child-trees is done.
			 */
			virtual const QStringList childlist() const {
				QStringList tmp;
				for ( childlist_t::const_iterator it=_childs.begin(); it!=_childs.end(); ++it )
					tmp += ( *it ).first;
				return tmp;
			}

			/**
			 * @brief Return the path from the root to this item
			 */
			virtual QString path() const;

			/**
			 * @brief Get a named child by traversing the tree
			 */
			AbstractTreeItem* getNamedChild( QString ) const;

			/**
			 * @brief Get a named child of type T by traversing the tree
			 */
			template<class T>
			T getNamedChild( QString n ) const { return dynamic_cast<T>( getNamedChild( n ) ); }

			/**
			 * @brief Get a named direct child
			 */
			AbstractTreeItem* child( QString ) const;

			/**
			 * @brief Get a named child of type T
			 */
			template<class T>
			T child( QString n ) const { return dynamic_cast<T>( child( n ) ); }

			/**
			 * @brief Dump a tree of this item and its childs for debug
			 */
			void debugPrint( QString prefix="" ) const {
				if ( prefix == "" )
					qDebug() << "AbstractTreeItem::debugPrint()";
				qDebug( "%s", QString( "%1 +%2 (childs %3)" ).arg( prefix ).arg( debug() ).arg( childs() ).toStdString().c_str() );
				for ( childlist_t::const_iterator it=_childs.begin(); it!=_childs.end(); ++it )
					( *it ).second->debugPrint( prefix+" |" );
			}

		protected:
			/**
			 * @brief Return debug-info for this item
			 *
			 * This has to be implemented by the implementations to print class-
			 * specific information in debugPrint().
			 */
			virtual QString debug() const =0;

			/**
			 * @brief The childs of this item
			 */
			childlist_t _childs;
			/**
			 * @brief The parent of this item
			 */
			AbstractTreeItem* _parent;

		private:
			/**
			 * @brief The name of this item
			 */
			QString _name;
	};

};

#endif // ABSTRACTTREEITEM_H

