/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DATATYPES_H
#define DATATYPES_H

#include <QtCore/QVariant>

#include <abstracttreeitem.h>
#include <namedqvarianttreeitem.h>

namespace NSCore {

	/**
	 * @brief Next-generation datasets.
	 *
	 * Should be more flexible than AcqisitionDataSet/ChannelDataSet
	 *
	 * This has
	 * <ul>
	 * <li>All the child elements in one tree
	 * <li>No distinction between parameters and results
	 * <li>Prefixes for the results reflecting the evaluation done on the data
	 * <li>Data-generations that hold modifications of the data
	 * <li>Generations and prefixes per dataset, not per channel
	 * <li>Different generations of the data can have different numbers of channels and samples.
	 * </ul>
	 */
	class AcqisitionDataSet
	{
		public:

			/**
			 * @brief Create a new dataset with the given number of channels and
			 * the given number of samples per channel in the first data-generation
			 */
			AcqisitionDataSet( int channels=0, int samples=0 );

			/**
			 * @brief Copy-constructor
			 */
			AcqisitionDataSet( const AcqisitionDataSet& );

			/// destructor
			~AcqisitionDataSet();

			/**
			 * @{
			 *
			 * @brief Add and remove a prefix from the tree of results/generations
			 *
			 * pushPrefix() does not create a new generation of data! You have
			 * to use newGeneration() for that.
			 *
			 * But nevertheless it makes sense to have these functions for
			 * example to add the fact that the data was filtered or that it was
			 * changed without creation of a new generation...
			 */
			void pushPrefix( QString );
			void popPrefix( QString );
			// @}

			/**
			 * @brief Return the current data generation
			 */
			int currentGenerationNumber() const;

			/**
			 * @brief Return the generations parent generation.
			 *
			 * -1 means the current generation
			 */
			int parentGeneration( int generation = -1 ) const;

			/**
			 * @brief create a new named generation of data
			 *
			 * Create a new generation of raw-data (with the given number of
			 * channels and samples) and create a new prefix that has the new
			 * generation attached.
			 */
			void newGeneration( QString prefix, int channels, int samples );

			/**
			 * @brief Return the prefix for a specific generation
			 */
			QString getPrefix( int generation = -1 ) const;

			/**
			 * @brief Add a result / parameter to the results tree
			 */
			void addResult( QString name, QVariant value );
			/**
			 * @brief Get a result
			 */
			NamedQVariantTreeItem* result( QString name );

			/**
			 * @brief Access the whole result-tree
			 */
			NamedQVariantTreeItem* resulttree();

			/**
			 * @{
			 *
			 * @brief access and manipulate the data sample-by-sample
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			double sample( int channel, int sample, int generation=-1 ) const __attribute__( ( deprecated ) );
			void setSample( int channel, int sample, double value, int generation=-1 ) __attribute__( ( deprecated ) );
			// @}

			/**
			 * @brief Access the raw double-array for one channel
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			double* getRawChannel( int channel, int generation=-1 );
			/**
			 * @brief Access the raw double-array for all channels
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			double* getRawData( int generation=-1 );
			/**
			 * @brief Update min/max after setting data via the getRaw*-functions
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 *
			 * @note This is pretty slow. Only use this after you really changed
			 * the data via getRaw*.
			 */
			void updateMinMax( int generation=-1 );
			/**
			 * @brief Reset the min/max-cache for the given generation
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			void resetMinMaxCache( int generation=-1 );

			/**
			 * @brief Return the number of channels in the given generation
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			int channels( int generation=-1 ) const;
			/**
			 * @brief Return the number of samples in the given generation
			 *
			 * A generation of -1 ( the default-argument ) means to use the
			 * current generation.
			 */
			int samples( int generation=-1 ) const;


			/// Just for debug
			void debugPrint() const;

		private:
			struct AcqisitionDataSet_Private;
			AcqisitionDataSet_Private* d;
	};

};

#endif // DATATYPES_H

