/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "diskiobuffers.h"
#include "diskiobuffers.moc"

#include <cmath>

#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>

#include "mathexpression.h"

namespace NSDiskIO {

	AcqirisLiveReader::AcqirisLiveReader( QObject* p )
		: AbstractReader( p )
		, AbstractDiskIO<QTextStream>( QIODevice::ReadOnly )
		, _samplinginterval( 0 )
		, _channelfsr( 5 )
		, _fullscale( 256 )
		, _horizontalposition( 0 )
		, _lsb( 0 )
	{
		qDebug() << "AcqirisLiveReader::AcqirisLiveReader(" << p << ")";
	}

	void AcqirisLiveReader::readSingleDataSet() {
		if ( _stream != 0 ) {
			bool finished = false;
			int i=0;
			while ( ! _stream->atEnd() && !finished ) {
				QString line = _stream->readLine();
				if ( line[ 0 ] == '$' ) {
					QString first = line.section( " ", 0, 0 );
					QString last = line.section( " ", -1, -1 );
					if ( first == "$SAMPIVAL" )
						_samplinginterval = last.toDouble();
					if ( first == "$CHANNELFSR" )
						_channelfsr = last.toDouble();
					if ( first == "$FULLSCALE" )
						_fullscale = last.toDouble();
					if ( first == "$HORPOS" )
						_nexthorizontalposition = last.toDouble();

					if ( _horizontalposition == 0 )
						_horizontalposition = _nexthorizontalposition;

					_lsb = _channelfsr / _fullscale;

					// only stop If you have read some data:
					if ( _data.size() > 0 )
						finished = true;
				} else {
					++i;
					_data.push_back( line.toDouble() * _lsb );
				}
			}

			NSCore::AcqisitionDataSet dataset( 1, _data.size() );
			dataset.addResult( "horizontalposition", _horizontalposition );
			dataset.addResult( "samplinginterval", _samplinginterval );
			dataset.addResult( "channel 0/fullscale", _channelfsr );

			double* ch = dataset.getRawData();
			for ( int i=0; i<_data.size(); ++i ) {
				ch[ i ] = _data[ i ];
			}
			dataset.updateMinMax();

			enqueue( dataset );

			_data.clear();
			_horizontalposition = _nexthorizontalposition;
		}
	}

	CSVWriter::CSVWriter( QObject* p ) : AbstractWriter( p ), AbstractDiskIO<QTextStream>( QIODevice::WriteOnly ) {
		setObjectName( "CSVWriter" );
		config[ "Filename to write to" ]->setUrlFilter( "*.csv" );
		_timealign = new NSCore::MathTreeRoot( this );
		_timealign->setObjectName( "Align data to" );
		QDomDocument doc("xml");
		doc.setContent( QString( "<path name=\"Align data to\"><add><result name=\"horizontalposition\" /><result name=\"triggerdelay\" /></add></path>" ) );
		_timealign->fromXml( doc.documentElement() );
		_timeinterval = new NSCore::MathTreeRoot( this );
		_timeinterval->setObjectName( "Time interval to use" );
		doc.setContent( QString( "<path name=\"Time interval to use\"><result name=\"samplinginterval\" /></path>" ) );
		_timeinterval->fromXml( doc.documentElement() );
		config->addParameter( "Generations to save", "0" );
		config[ "Generations to save" ]->setToolTip( "Comma-separated list of generations to save to the file. Data format will be: time, <list of all channels for each given generation>" );
		//debugPrint();
	}
	void CSVWriter::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {

			QList<int> generations;
			foreach( QString s, config[ "Generations to save" ]->toString().split( ",", QString::SkipEmptyParts ) )
				generations.append( s.toInt() );

			double samplinginterval = _timeinterval->value( data );
			if ( isnan( samplinginterval ) || samplinginterval == 0 )
				samplinginterval = 1;

			double horizontalposition = _timealign->value( data );
			if ( isnan( horizontalposition ) )
				horizontalposition = 0;

			for ( int i=0; i<data.samples(); ++i ) {
				*_stream << horizontalposition + i*samplinginterval;
				foreach( int gen, generations )
					for ( int j=0; j<data.channels( gen ); ++j ) {
						double* ch = data.getRawChannel( j, gen );
						if ( i < data.samples( gen ) )
							*_stream << ", " << ch[ i ];
					}
				*_stream << endl;
			}

			*_stream << endl << endl;

		}
	}

	EPOSWriter::EPOSWriter( QObject* p ) : AbstractWriter( p ), AbstractDiskIO<QTextStream>( QIODevice::WriteOnly ) { }
	void EPOSWriter::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {
			*_stream << data.result( "samplinginterval" )->toDouble()
			         << " "
			         << data.result( "triggerdelay" )->toDouble()
			         << " "
			         << data.result( "horizontalposition" )->toDouble()
			         << " "
			         << data.result( "timestamp" )->toDouble()
			         << " "
			         << data.channels()
			         << " "
			         << data.samples()
			         << " ";
			for ( int i=0; i<data.channels(); ++i ) {
				*_stream << data.result( QString( "channel %1/fullscale" ).arg( i ) )->toDouble()
				         << " "
				         << data.result( QString( "channel %1/offset" ).arg( i ) )->toDouble()
				         << " "
				         ;
				double* ch = data.getRawChannel( i );
				for ( int j=0; j<data.samples(); ++j )
					*_stream << ch[ j ] << " ";
			}
			*_stream << endl;
		}
	}

	EPOSReader::EPOSReader( QObject* p ) : AbstractReader( p ), AbstractDiskIO<QTextStream>( QIODevice::ReadOnly ) { }
	void EPOSReader::readSingleDataSet() {
		if ( _stream != 0 && ! _stream->atEnd() ) {

			double samplinginterval, triggerdelay, horizontalposition, timestamp;
			int channels, samples;
			*_stream >> samplinginterval; _stream->skipWhiteSpace();
			*_stream >> triggerdelay; _stream->skipWhiteSpace();
			*_stream >> horizontalposition; _stream->skipWhiteSpace();
			*_stream >> timestamp; _stream->skipWhiteSpace();
			*_stream >> channels; _stream->skipWhiteSpace();
			*_stream >> samples; _stream->skipWhiteSpace();

			NSCore::AcqisitionDataSet data( channels, samples );
			data.addResult( "samplinginterval", samplinginterval );
			data.addResult( "triggerdelay", triggerdelay );
			data.addResult( "horizontalposition", horizontalposition );
			data.addResult( "timestamp", timestamp );

			double fullscale, offset, sample;
			for ( int i=0; i<channels; ++i ) {
				*_stream >> fullscale; _stream->skipWhiteSpace();
				*_stream >> offset; _stream->skipWhiteSpace();
				data.addResult( QString( "channel %1/fullscale" ).arg( i ), fullscale );
				data.addResult( QString( "channel %1/offset" ).arg( i ), offset );

				double* ch = data.getRawChannel( i );
				for ( int j=0; j<samples; ++j ) {
					*_stream >> sample; _stream->skipWhiteSpace();
					ch [ j ] = sample;
				}

			}
			data.updateMinMax();

			enqueue( data );
		}
	}
	void EPOSReader::setFile( QString file ) {
		AbstractDiskIO<QTextStream>::setFile( file );
	}
	QString EPOSReader::filename() const {
		return AbstractDiskIO<QTextStream>::filename();
	}

	CombinedCSV::CombinedCSV( QObject* p )
		: AbstractWriter( p )
		, AbstractDiskIO<QTextStream>( QIODevice::WriteOnly )
		, _combine_counter( 0 )
	{
		setObjectName( "CombinedCSV" );
		config[ "Filename to write to" ]->setUrlFilter( "*.csv" );
		config->addParameter( "Datasets to average over", 1 );
		config[ "Datasets to average over" ]->setRange( 1, int( 1E8 ) );
		config->addParameter( "Generation", QVariant::Int );
	}

	void CombinedCSV::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {
			int generation = config[ "Generation" ]->toInt();

			for ( int c=0; c<data.channels( generation ); ++c ) {
				double* ch = data.getRawChannel( c, generation );
				while ( _save[ c ].size() < data.samples( generation ) )
					_save[ c ].append( 0.0 );
				for ( int i=0; i<data.samples( generation ); ++i ) {
					_save[ c ][ i ] += ch[ i ];
				}
			}
			_combine_counter++;

			if ( _combine_counter >= config[ "Datasets to average over" ]->toInt() ) {
				qDebug() << " Will save" << _combine_counter << "averaged datasets to disk.";

				for ( int i=0; i<data.samples( generation ); ++i ) {

					*_stream << i;
					foreach( int c, _save.keys() ) {
						*_stream << " " << _save[ c ][ i ] / _combine_counter;
					}
					*_stream << endl;
				}
				*_stream << endl << endl;

				_save.clear();
				_combine_counter = 0;
			}

		}
	}
}

