/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef EVALUATIONPATH_ITEMS_H
#define EVALUATIONPATH_ITEMS_H

#include <QtCore/QObject>
#include <QtCore/QDebug>

#include "evaluationpath.h"
#include "plugins_factory.h"


namespace NSEvaluation {

	class AnalyzerItem;
	class ModifierItem;
	class FilterItem;

	/**
	 * @brief The base for the evaluationchain
	 *
	 * This class is the base for a tree to do evaluation on the datasets.
	 */
	class PluginItem : public NSCore::QTreeItem {
		Q_OBJECT

		protected:
			/// constructor
			PluginItem( PluginItem* p=0 ) : QTreeItem( p )
			{ }
			/// constructor
			PluginItem( QString n, QTreeItem* p ) : QTreeItem( n, p )
			{ }
		public:
			/// destructor
			~PluginItem() {
			}

			/**
			 * @brief process data
			 *
			 * This function is used to process the data. This implementation
			 * does nothing, derived class have to to the real thing.
			 */
			virtual void process( NSCore::AcqisitionDataSet& ) {
				qDebug() << "PluginItem::process( data )";
			}

			// TreeItem
			virtual QString type() const =0;
			virtual QString name() const =0;
			virtual QString help() const { return "<qt><i>Sorry, no help set.</i></qt>"; }

			/**
			 * @brief Give access to the configuration
			 */
			virtual NSCore::NamedParameterTreeItemPointer config() =0;

			QStringList guiActions() const {
				return QTreeItem::guiActions() << "Add" << "Remove";
			}
			void guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
				if ( action == "Add" )
					emit sExecAdd( this, gui );
				else if ( action == "Remove" )
					emit sExecRemove( this, gui );
				else
					NSCore::QTreeItem::guiExecAction( action, gui );
			}
			QString guiString() const { return QString( "%1: %2" ).arg( type() ).arg( name() ); }

		signals:
			void sExecAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* );
			void sExecRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* );

		protected:
			/// debug...
			QString debug() const { return QString( "%1(%2)" ).arg( type() ).arg( name() ); }

			bool createFromXML( QDomNode node );

	};

	/**
	 * @brief Protected pointer to a PluginItem
	 */
	class PluginItemPointer : public QObject, public NSCore::Pointer_template<PluginItem>
	{
		Q_OBJECT
		public:
			PluginItemPointer( PluginItem* item, QObject* p ) : QObject( p ), NSCore::Pointer_template<PluginItem>( item )
			{ }


			void setNull() {
				qDebug() << "PluginItemPointer::setNull()";
				if ( this != 0 && _p != 0 ) {
					emit deleted( _p );
					emit deleted();
				}
				NSCore::Pointer_template<PluginItem>::setNull();
			}

		signals:
			void deleted( PluginItem* );
			void deleted();

	};

	/**
	 * @internal to EvaluationPath
	 */
	class AnalyzerItem : public PluginItem
	{
		public:
			AnalyzerItem( QString plugin, QTreeItem* p )
				: PluginItem( plugin, p )
				, _analyzer( dynamic_cast<Analyzer*>( PluginFactory::the()->newPlugin( plugin, this ) )  )
			{
				// The one allowed child is named options and is a NamedParameterTreeItem :-)
				//maximumChilds( 1 );
				setObjectName( _analyzer->type() );
			}
			~AnalyzerItem() {
				delete _analyzer;
			}

			Analyzer* analyzer() { return _analyzer; }

			void process( NSCore::AcqisitionDataSet& data ) {
				if ( _analyzer )
					_analyzer->analyze( data );
			}

			QString type() const { return "Analyzer"; }
			QString name() const {
				if ( _analyzer )
					return QString( "%1" ).arg( _analyzer->type() );
				return "(No Analyzer found/set!)";
			}
			QString help() const {
				if ( _analyzer )
					return _analyzer->help();
				return "";
			}

			NSCore::NamedParameterTreeItemPointer config() { return _analyzer->config(); }

			QDomNode saveToXML( QDomDocument doc ) {
				QDomElement node = doc.createElement( "analyzer" );
				node.setAttribute( "type", _analyzer->type() );
				return node;
			}
			bool restoreFromXML( QDomNode node ) { return node.nodeName() == "analyzer"; }

			QStringList guiActions() const {
				return QTreeItem::guiActions() << "Remove";
			}
		private:
			Analyzer* _analyzer;
	};

	/**
	 * @internal to EvaluationPath
	 */
	class ModifierItem : public PluginItem
	{
		public:
			ModifierItem( QString plugin, QTreeItem* p )
				: PluginItem( plugin, p )
				, _config( new NSCore::NamedParameterTreeItem( "options", QVariant::Invalid, this ) )
				, _modifier( dynamic_cast<Modifier*>( PluginFactory::the()->newPlugin( plugin, _config ) ) )
			{
				_config->addParameter( "new generation", QVariant::Bool );
				_config->getParameter( "new generation" )->setValue( true );
				setObjectName( _modifier->type() );
			}
			~ModifierItem() {
				delete _modifier;
			}
			Modifier* modifier() { return _modifier; }

			void process( NSCore::AcqisitionDataSet& data ) {
				bool newgeneration = _config->getParameter( "new generation" )->toBool();

				int oldgeneration = data.currentGenerationNumber();

				_modifier->prepare( data );

				if ( _modifier->channels > data.channels() || _modifier->samples > data.samples() )
					newgeneration = true;

				if ( newgeneration )
					data.newGeneration( _modifier->prefix, _modifier->channels, _modifier->samples );
				else
					data.pushPrefix( _modifier->prefix );

				data.resetMinMaxCache();

				if ( _modifier != 0 )
					_modifier->modify( data, oldgeneration );

				foreach( QObject* abstractitem, childobjects() ) {
					PluginItem* item = dynamic_cast<PluginItem*>( abstractitem );
					if ( item != 0 )
						item->process( data );
				}

				data.popPrefix( _modifier->type() );
			}

			QString type() const { return "Modifier"; }
			QString name() const { if ( _modifier != 0 ) return _modifier->type(); return "(Empty)"; }
			QString help() const {
				if ( _modifier )
					return _modifier->help();
				return "";
			}
			NSCore::NamedParameterTreeItemPointer config() { return _config; }

			QDomNode saveToXML( QDomDocument doc ) {
				QDomElement node = doc.createElement( "modifier" );
				node.setAttribute( "type", _modifier->type() );
				return node;
			}
			bool restoreFromXML( QDomNode node ) { return node.nodeName() == "modifier"; }
		private:
			NSCore::NamedParameterTreeItem* _config;
			Modifier* _modifier;
	};

	/**
	 * @internal to EvaluationPath
	 */
	class FilterItem : public PluginItem
	{
		public:
			FilterItem( QString plugin, QTreeItem* p )
				: PluginItem( plugin, p )
				, _config( new NSCore::NamedParameterTreeItem( "options", QVariant::Invalid, this ) )
				, _filter( dynamic_cast<Filter*>( PluginFactory::the()->newPlugin( plugin, _config ) ) )
			{
				_config->addParameter( "good pulses", QVariant::Int );
				_config->getParameter( "good pulses" )->setMin( 0 );
				_config->getParameter( "good pulses" )->setMax( int( 1E10 ) );
				_config->getParameter( "good pulses" )->setGuiHint( NSCore::NamedParameterTreeItem::Counter );
				setObjectName( _filter->type() );
			}
			~FilterItem() {
				delete _filter;
			}

			void process( NSCore::AcqisitionDataSet& data ) {
				bool go = false;
				if ( _filter )
					go = _filter->filter( data );

				if ( go ) {
					_config->getParameter( "good pulses" )->setValue( _config->getParameter( "good pulses" )->toInt() + 1 );
					foreach( QObject* abstractitem, childobjects() ) {
						PluginItem* item = dynamic_cast<PluginItem*>( abstractitem );
						if ( item != 0 )
							item->process( data );
					}
				}
			}

			QString type() const { return "Filter"; }
			QString name() const {
				if ( _filter )
					return QString( "%1" ).arg( _filter->type() );
				return "(No Filter found/set!)";
			}
			QString help() const {
				if ( _filter )
					return _filter->help();
				return "";
			}

			NSCore::NamedParameterTreeItemPointer config() { return _config; }

			QDomNode saveToXML( QDomDocument doc ) {
				QDomElement node = doc.createElement( "filter" );
				node.setAttribute( "type", _filter->type() );
				return node;
			}
			bool restoreFromXML( QDomNode node ) { return node.nodeName() == "filter"; }
		private:
			NSCore::NamedParameterTreeItem* _config;
			Filter* _filter;
	};

};

#endif // EVALUATIONPATH_ITEMS_H

