/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "interpolation_helpers.h"
#include <math.h>
#include <QtCore/QDebug>

double PolynomInterpolate::polint( double x ) {
	int n = _order;
	int i, m, ns=1;
	double den, dif, dift, ho, hp, w;
	double c[ n ], d[ n ];
	double y=0, dy=0;

	dif = fabs( x - 0 );

	for ( i=1; i<=n; i++ ) {
		if ( ( dift = fabs( x - i-1 ) ) < dif ) {
			ns = i;
			dif = dift;
		}
		c[ i ] = _ya[ i-1 + _begin ];
		d[ i ] = _ya[ i-1 + _begin ];
	}

	y = _ya[ ( ns-- ) -1 + _begin ];
	for ( m=1; m<n; m++ ) {
		for ( i=1; i<=n-m; i++ ) {
			ho = i-1 + _begin - x;
			hp = i+m-1 + _begin - x;
			w = c[ i+1 ] - d[ i ];
			if ( ( den=ho-hp ) == 0.0 )
				qFatal( "Error in routine polint" );
			den = w/den;
			d[ i ] = hp * den;
			c[ i ] = ho * den;
		}
		y += ( dy=( 2*ns < ( n-m ) ? c[ ns+1 ] : d[ ns-- ] ) );
	}

	return y;
}

#define R 0.61803399
#define C ( 1.0-R )
#define SHFT2( a,b,c ) ( a )=( b ); ( b )=( c );
#define SHFT3( a,b,c,d ) ( a )=( b ); ( b )=( c ); ( c )=( d );

coord Minimum::golden_minimum() {
	coord ret;

	double ax = _minimum_at_int -1, bx = _minimum_at_int, cx = _minimum_at_int+1;
	double tol = 1E-7;
	double f1,f2,x0,x1,x2,x3;

	x0 = ax;
	x3 = cx;
	if ( fabs( cx-bx ) > fabs( bx-ax ) ) {
		x1 = bx;
		x2 = bx + C * ( bx-ax );
	} else {
		x2 = bx;
		x1 = bx - C * ( bx-ax );
	}
	f1 = function( x1 );
	f2 = function( x2 );
	while ( fabs( x3-x0 ) > tol * ( fabs( x1 )+fabs( x2 ) ) ) {
		if ( f2 < f1 ) {
			SHFT3( x0,x1,x2,R*x1+C*x3 )
			SHFT2( f1,f2,function( x2 ) )
		} else {
			SHFT3( x3,x2,x1,R*x2+C*x0 )
			SHFT2( f2,f1,function( x1 ) )
		}
	}
	if ( f1<f2 ) {
		ret.x = x1;
		ret.y = f1;
	} else {
		ret.x = x2;
		ret.y = f2;
	}

	return ret;
}

coord Maximum::golden_maximum() {
	coord ret;

	double ax = _maximum_at_int -1, bx = _maximum_at_int, cx = _maximum_at_int+1;
	double tol = 1E-7;
	double f1,f2,x0,x1,x2,x3;

	x0 = ax;
	x3 = cx;
	if ( fabs( cx-bx ) < fabs( bx-ax ) ) {
		x1 = bx;
		x2 = bx + C * ( bx-ax );
	} else {
		x2 = bx;
		x1 = bx - C * ( bx-ax );
	}
	f1 = function( x1 );
	f2 = function( x2 );
	while ( fabs( x3-x0 ) > tol * ( fabs( x1 )+fabs( x2 ) ) ) {
		if ( f2 > f1 ) {
			SHFT3( x0,x1,x2,R*x1+C*x3 )
			SHFT2( f1,f2,function( x2 ) )
		} else {
			SHFT3( x3,x2,x1,R*x2+C*x0 )
			SHFT2( f2,f1,function( x1 ) )
		}
	}
	if ( f1>f2 ) {
		ret.x = x1;
		ret.y = f1;
	} else {
		ret.x = x2;
		ret.y = f2;
	}

	return ret;
}

Spline::Spline()
	: _y( 0 )
	, _n( 0 )
	, _y2( 0 )
	, _yp1( 0 )
	, _ypn( 0 )
	, _u( 0 )
{
}
Spline::Spline( double* y, int n, double yp1, double ypn )
	: _y( 0 )
	, _n( 0 )
	, _y2( 0 )
	, _yp1( yp1 )
	, _ypn( ypn )
	, _u( 0 )
{
	qDebug() << "Spline::Spline(" << y << "," << n << "," << yp1 << "," << ypn << ")";
	setData( y, n );
	//createTable();
}
Spline::~Spline() {
	_y = 0;
	delete _y2;
	delete _u;
}

void Spline::setData( double* y, int n ) {
	//qDebug() << "Spline::setData(" << y << "," << n << ")";
	//if ( y != _y ) {
		_y = y;
		if ( n != _n ) {
			delete _y2;
			_y2 = new double[ n ];
			delete _u;
			_u = new double[ n ];
			//_u = 0;
			_n = n;
		}
		createTable();
	//}
}

void Spline::createTable() {
	qDebug() << "Spline::createTable()";
	double p, qn, sig, un;

	if ( _u == 0 )
		_u = new double[ _n ];

	if ( _yp1 > 0.99e30 )
		_y2[ 0 ] = _u[ 0 ] = 0.0;
	else {
		_y2[ 0 ] = -0.5;
		_u[ 0 ] = 3.0*( ( y( 1 )-y( 0 ) )-_yp1 );
	}

	qDebug() << "1";
	for ( int i=1; i<_n-1; ++i ) {
		sig = 1/2;
		p = sig*_y2[ i-1 ]+2.0;
		_y2[ i ] = ( sig-1.0 )/p;
		_u[ i ] = ( y( i+1 )-y( i ) ) - ( y( i )-y( i-1 ) );
		_u[ i ] = ( 6.0*_u[ i ]/2 - sig*_u[ i-1 ] )/p;
	}

	qDebug() << "2";
	if ( _ypn > 0.99e30 )
		qn = un = 0.0;
	else {
		qn = 0.5;
		un = 3.0*( _ypn-( y( _n )-y( _n-1 ) ) );
	}

	qDebug() << "3";
	_y2[ _n ] = ( un-qn*_u[ _n-1 ] )/( qn*_y2[ _n-1 ]+1.0 );
	for ( int k=_n-2; k>=0; --k )
		_y2[ k ] = _y2[ k ]*_y2[ k+1 ]+_u[ k ];

	qDebug() << "Spline::createTable() finished";
}

double Spline::splint( double x ) {
	//qDebug() << "Spline::splint(" << x << ")";
	if ( ! _y ) {
		qWarning() << "Spline::splint() : No _y, will return 0!";
		return 0.0;
	}
	if ( _u == 0 )
		createTable();

	int klo, khi, k;
	double h, b, a;

	klo = 0;
	khi = _n-1;
	while ( ( khi-klo ) > 1 ) {
		k = ( khi+klo ) >> 1;
		if ( k > x ) khi = k;
		else klo = k;
	}

	h = khi*1.0-klo*1.0;
	if ( h == 0.0 ) {
		qWarning() << " h == 0.0 !";
		return 0;
	}

	a = ( khi*1.0-x )/h;
	b = ( x-klo*1.0 )/h;

	return a*y( klo ) + b*y( khi ) + ( ( a*a*a-a )*_y2[ klo ] + ( b*b*b-b )*_y2[ khi ] )*( h*h )/6.0;
}

void Spline::debug() const {
	for ( int i=0; i<_n; ++i ) {
		qDebug() << i << _y2[ i ];
	}
}

