/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "modifiers.h"
//#include "modifiers.moc"

#include <math.h>
#include <QtCore/QDebug>
#include <analyzers.h>

AbsModifier::AbsModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void AbsModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	//qDebug() << "AbsModifier::modify( data )";
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		for ( int j=0; j<data.samples(); ++j )
			out[ j ] = fabs( in[ j ] );
	}
	data.updateMinMax();
}

NormalizeModifier::NormalizeModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	for ( int i=0; i<2; ++i ) {
		_config->addParameter( QString( "Channel %1/Height to Normalize" ).arg( i ), 1.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMax( 100.0 );
		_config->addParameter( QString( "Channel %1/Offset after Normalize" ).arg( i ), 0.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMax( 100.0 );
	}
}

void NormalizeModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );

		double height_to_normalize = _config[ QString( "Channel %1/Height to Normalize" ).arg( i%2 ) ]->toDouble();
		double offset_after_normalize = _config[ QString( "Channel %1/Offset after Normalize" ).arg( i%2 ) ]->toDouble();

		double factor = 1/height_to_normalize;

		for ( int j=0; j<data.samples(); ++j )
			out[ j ] = in[ j ] * factor + offset_after_normalize;

		data.addResult( QString( "Normalized by factor (Channel %1)" ).arg( i ), factor );
		data.addResult( QString( "Added Offset (Channel %1)" ).arg( i ), offset_after_normalize );

	}
	data.updateMinMax();
}

void simpleAverager( int neighbours, int samplesize, const double* in, double* out ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];

	for ( int i=0; i<vsize; ++i )
		v[ i ] = in[ i ];

	for ( int i=0; i<neighbours; ++i ) {
		out[ i ] = in[ i ];
		out[ samplesize - i - 1 ] = in[ samplesize - i - 1 ];
	}

	for ( int i=neighbours; i<samplesize-neighbours; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += v[ j ];
		value /= vsize;

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		v[ vsize-1 ] = in[ i + neighbours + 1 ];
	}
}
void simpleAveragerV2( int neighbours, int samplesize, const double* in, double* out ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];

	for ( int i=0; i<neighbours; ++i )
		v[ i ] = 0;
	for ( int i=neighbours; i<vsize; ++i )
		v[ i ] = in[ i-neighbours ];

	for ( int i=0; i<samplesize; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += v[ j ];
		value /= vsize;

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		if ( i<samplesize-neighbours )
			v[ vsize-1 ] = in[ i + neighbours + 1 ];
		else
			v[ vsize-1 ] = 0.0;
	}
}
void weightedAverager( int neighbours, int samplesize, const double* in, double* out, double* weights ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];
	if ( weights == 0 ) {
		weights = new double[ vsize ];
		for ( int i=0; i<vsize; ++i )
			weights[ i ] = 1.0/vsize;
	}

	for ( int i=0; i<neighbours; ++i )
		v[ i ] = 0;
	for ( int i=neighbours; i<vsize; ++i )
		v[ i ] = in[ i-neighbours ];

	for ( int i=0; i<samplesize; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += weights[ j ] * v[ j ];

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		if ( i<samplesize-neighbours )
			v[ vsize-1 ] = in[ i + neighbours + 1 ];
		else
			v[ vsize-1 ] = 0.0;
	}
}

FloatingAverageModifier::FloatingAverageModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	qDebug() << "FloatingAverageModifier::FloatingAverageModifier()";
	_config->addParameter( "neighbours", 1 );
	_config[ "neighbours" ]->setMin( 0 );
	_config[ "neighbours" ]->setMax( 10 );
	_config[ "neighbours" ]->setToolTip( "The number of neighbours to the left and right that is looked at." );
}

void FloatingAverageModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	int neighbours = _config[ "neighbours" ]->toInt();
	for ( int i=0; i<data.channels(); ++i ) {

		double *_in = data.getRawChannel( i, oldgeneration );
		double *_out = data.getRawChannel( i );

		int size = neighbours * 2 + 1;
		double *weights = new double[ size ];
		double s = 0.75;
		double sum = 0;
		for ( int i=0; i<size; i++ ) {
			weights[ i ] = exp( - pow( i-neighbours, 2 ) / ( 2 * pow( s, 2 ) ) );
			sum += weights[  i ];
		}
		for ( int i=0; i<size; ++i )
			weights[ i ] /= sum;
		weightedAverager( neighbours, data.samples(), _in, _out, weights );
		delete weights;
	}
	data.updateMinMax();
}

IntegrateModifier::IntegrateModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void IntegrateModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		double old = 0;
		for ( int j=0; j<data.samples(); ++j ) {
			old += in[ j ];
			out[ j ] = old;
		}
	}
}

PulseIntegrate::PulseIntegrate( NSCore::QTreeItem* p )
	: NSEvaluation::Modifier( p )
{
	_config->addParameter( "samples before", 20 );
	_config->addParameter( "samples after", 30 );
}
void PulseIntegrate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	int before = _config[ "samples before" ]->toInt();
	int after = _config[ "samples after" ]->toInt();
	int start = data.samples();
	int stop = 0;
	for ( int i=0; i<data.channels(); ++i ) {
		start = qMin( start, data.result( QString( "../channel %1/min at" ).arg( i ) )->toInt() - before );
		stop = qMax( stop, data.result( QString( "../channel %1/min at" ).arg( i ) )->toInt() + after );
	}
	//start = qMax( start, 0 );
	//stop = qMin( stop, data.samples() );
	if ( start>=0 && stop< data.samples() )
		for ( int i=0; i<data.channels(); ++i ) {
			double* in = data.getRawChannel( i, oldgeneration );
			double* out = data.getRawChannel( i );
			if ( in != out ) {
				memcpy( out, in, sizeof( double ) * data.samples() );
			}
			for ( int j = start; j<stop; ++j ) {
				out[j] = out[j-1] + in[j];
			}
			for ( int j=stop; j<data.samples(); ++j ) {
				out[j] = out[j-1] - in[j];
			}
		}
	else
		for ( int i=0; i<data.channels(); ++i ) {
			double * out = data.getRawChannel( i );
			for ( int j=0; j<data.samples(); ++j )
				out[ j ] = 0.0;
		}
	data.updateMinMax();
}

SignedIntegrate::SignedIntegrate( NSCore::QTreeItem* p )
	:NSEvaluation::Modifier( p )
{
}

void SignedIntegrate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		double last = 0; //data.channel( i )->sample( 0, 0 );
		double now = 0, sum = 0;
		//data.channel( i )->setSample( 0, sum );
		for ( int j=0; j<data.samples(); ++j ) {

			now = in[ j ];

			if ( now < last )
				sum -= fabs( now );
			else
				sum += fabs( now );

			out[ j ] = sum;

			last = now;
		}
	}
}

Differentiate::Differentiate( NSCore::QTreeItem* p )
	: NSEvaluation::Modifier( p )
{
}
void Differentiate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* olddata = data.getRawChannel( i, oldgeneration );
		double* newdata = data.getRawChannel( i );

		double last = 0;
		for ( int j=0; j<data.samples(); ++j ) {
			double tmp = olddata[ j ];
			newdata[ j ] = tmp - last;
			last = tmp;
		}
	}
	data.updateMinMax();
}

CrossCorrelation::CrossCorrelation( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "middle percentage", 100 );
	_config[ "middle percentage" ]->setMin( 0 );
	_config[ "middle percentage" ]->setMax( 100 );
	_config[ "middle percentage" ]->setToolTip( "The percentage of data that is to be evaluated. Will be aligned in the middle of the dataset..." );
}

void CrossCorrelation::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	if ( data.channels() < 2 ) {
		qDebug() << "Dataset doesn't contain enough channels.";
		return;
	}

	double percent = _config->getParameter( "middle percentage" )->toInt() / 100.0;

	double firstamplitude = 1;
	double secondamplitude = 1;
	QVariant* tmp;
	if ( ( tmp = data.result( "../channel 0/interpolated minimum" ) )->isValid() )
		firstamplitude = tmp->toDouble();
	if ( ( tmp = data.result( "../channel 1/interpolated minimum" ) )->isValid() )
		secondamplitude = tmp->toDouble();

	int ibounds = data.samples() - int( data.samples() * percent );
	ibounds /= 2;

	double* in1 = data.getRawChannel( 0, oldgeneration );
	double* in2 = data.getRawChannel( 1, oldgeneration );
	double* out = data.getRawChannel( 0 );
	for ( int i=0; i<ibounds; ++i )
		out[ i ] = 2;

	for ( int i=ibounds; i< data.samples() - ibounds; ++i ) {

		int joffset = qMax( data.samples() / 2 - i, 0 );
		int ioffset = qMax( i - data.samples()/2, 0 );
		int jend = data.samples()/2 + i;
		if ( jend > data.samples() )
			jend = data.samples()/2 + data.samples() - i;

		double sum = 0;
		double tmp;
		for ( int j=0; j<jend; ++j ) {
			tmp = in1[ ioffset + j ]/firstamplitude - in2[ joffset + j ]/secondamplitude;
			sum += tmp*tmp;
		}

		out[ i ] = sqrt( sum / jend );
	}

	for ( int i=data.samples() - ibounds; i<data.samples(); ++i )
		out[ i ] = 2;

}

