/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef SCRIPTENVIRONMENT_H
#define SCRIPTENVIRONMENT_H

#include <QtCore/QObject>
#include <QtCore/QStringList>
#include <QtCore/QMap>
#include <QtXml/QDomNode>

class QScriptEngine;
class QWidget;
class QSettings;

/**
 * @brief Scripting for EPOS!
 *
 * This namespace/module/lib contains stuff to use QScript for scripting-support of EPOS-apps.
 *
 * @todo dox!
 */
namespace NSScript {

	class ScriptSnippet;

	/**
	 * @brief The basic environment for scripting
	 *
	 * This provides an environment for scripting supporting the special EPOS-features.
	 *
	 * <ul>
	 * <li>Full access on all ( published ) QObjects. These are identified by their objectName.
	 * <li>Three predefined ( save- )places for scripts:
	 * <ol>
	 * <li>Inside the xml-file
	 * <li>As stand-alone .js-files
	 * <li>System-wide ( per user ) and restored on startup
	 * </ol>
	 * </ul>
	 */
	class ScriptEnvironment : public QObject
	{
		Q_OBJECT

		friend class ScriptIntrospection;

		public:
			/**
			 * @brief Construct a new ScriptEnvironment
			 */
			ScriptEnvironment( QObject* );
			/// destructor
			~ScriptEnvironment();

			/**
			 * @brief Add a QObject to the script-environment
			 *
			 * This adds a new QObject to the scripting environment and uses
			 * QObject::objectName() as the variable-name of the object. If no
			 * objectname is set, it will be named scriptobjectX where X is an
			 * increasing number.
			 */
			void addQObject( QObject* );

			/**
			 * @brief Wether there are unrestored ScriptSnippets after startup
			 */
			bool hasUnrestoredSnippets() const;
			/**
			 * @brief Restore the first unrestored ScriptSnippet
			 *
			 * @param w Parent widget of the new ScriptSnippet
			 */
			ScriptSnippet* restoreFirstSnippet( QWidget* w );

			/**
			 * @brief Create a new empty snippet
			 *
			 * @param w Parent widget of the new ScriptSnippet
			 */
			ScriptSnippet* newSnippet( QWidget* w );

			/**
			 * @brief Store the relevant snippets to xml
			 */
			QString toXML() const;

			/**
			 * @brief Restore snippets from xml
			 *
			 * @{
			 */
			void fromXML( QString );
			void fromXML( QDomNode );
			//@}

		signals:
			/**
			 * @brief There are snippets available to restore
			 */
			void snippetsToRestoreAvailable();
		private slots:
			/**
			 * @brief Execute the given code
			 */
			void execCode( QString );
			/**
			 * @brief A snippet was destroyed
			 */
			void snippetDestroyed( ScriptSnippet* );
			/**
			 * @brief Save the snippets that are saved system-wide
			 */
			void systemSave();

		private:

			/// Keep track of the elements in the scripting environment
			QObjectList _objects;

			/// The engine itself
			QScriptEngine* _engine;

			/// Active script snippets
			QList<ScriptSnippet*> _snippets;

			/// Settings-pointer
			QSettings* _settings;

			struct _snippetsToRestore_t {
				QString name;
				QString data;
				int /*ScriptSnippet::SaveLocation*/ savelocation;
			};
			/// List/Map of the snippets to restore after startup
			QList<_snippetsToRestore_t> _snippetsToRestore;
	};

};

#endif // SCRIPTENVIRONMENT_H

