/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "resulttospectrum.h"
#include "resulttospectrum.moc"

#include <QtCore/QDebug>
#include <QtCore/QTimer>
#include <QtCore/QThread>
#include <QtXml/QDomDocument>
#include <QtGui/QIcon>

#include <math.h>
#include <mathexpression.h>
#include "spectrummodel.h"
#include <thread_template.h>

ResultAction::ResultAction( ResultsToSpectrum* parent )
	: NSCore::MathTreeRoot( parent )
	, _model( 0 )
{
	//qDebug() << "ResultAction::ResultAction(" << parent << ")";
	setSingleChild( false );
}

ResultAction::~ResultAction() {
	//qDebug() << "ResultAction::~ResultAction()";
	if ( _model != 0 ) {
		delete _model;
		_model = 0;
	}
}

void ResultAction::data( NSCore::AcqisitionDataSet dataset ) {
	Abstract1DSpectrumModel* model1d =0;
	double x=NAN;
	double y=NAN;
	if ( _model != 0 )
		model1d = dynamic_cast<Abstract1DSpectrumModel*>( _model );
	foreach ( NSCore::MathExpression* exp, childobjects<NSCore::MathExpression>() ) {
		if ( exp ) {
			double ret = exp->value( dataset );
			if ( !isnan( ret ) && model1d )
				model1d->data( ret );
			if ( !isnan( ret ) ) {
				if ( isnan( x ) )
					x = ret;
				else {
					if ( isnan( y ) )
						y = ret;
				}
			}
		}
	}
	TwoDimSpectrumModel* model2d = dynamic_cast<TwoDimSpectrumModel*>( _model );
	if ( model2d && !isnan( x ) && !isnan( y ) )
		model2d->data( x,y );
}

void ResultAction::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	if ( action == "Remove Me" ) {
		if ( gui->messageYesNo( "Really?", "Do you really want to remove this spectrum with all its associated math?" ) )
			emit stopMe();
	} else
		NSCore::MathTreeRoot::guiExecAction( action, gui );
}

QDomNode ResultAction::saveToXML( QDomDocument doc ) {
	QDomElement node = doc.createElement( "spectrum" );
	node.setAttribute( "name", objectName() );
	node.setAttribute( "dimensions", _model->dimensions() );
	node.appendChild( _model->config()->toXml( doc ) );
	return node;
}
bool ResultAction::restoreFromXML( QDomNode node ) {
	if ( node.nodeName() != "spectrum" ) {
		return false;
	}
	QDomElement elem = node.toElement();
	setObjectName( elem.attributes().namedItem( "name" ).nodeValue() );
	int dimensions = 1;
	if ( elem.attributes().contains( "dimensions" ) ) {
		dimensions = elem.attributes().namedItem( "dimensions" ).nodeValue().toInt();
	}
	if ( dimensions == 2 )
		_model = new TwoDimSpectrumModel( this );
	else
		_model = new ClassicSpectrumModel( this );
	_model->setObjectName( objectName() );
	QDomElement path = elem.firstChildElement( "parameter" );
	if ( ! path.isNull() ) {
		_model->config()->fromXml( path );
	}

	return true;
}


Q_DECLARE_METATYPE( QModelIndex )
Q_DECLARE_METATYPE( NSCore::AcqisitionDataSet )

ResultsToSpectrum::ResultsToSpectrum( QObject* p )
	: NSCore::DataSetChainMember( p )
{
	qRegisterMetaType<QModelIndex>();
	qRegisterMetaType<NSCore::AcqisitionDataSet>();
	setObjectName( "Results2Spectrum" );
}
ResultsToSpectrum::~ResultsToSpectrum() {
}

void ResultsToSpectrum::removeAllPaths() {
	foreach( NSCore::QTreeItem* item, childobjects<NSCore::QTreeItem>() )
		item->deleteLater();
}

void ResultsToSpectrum::addPath( QString xml ) {
	qDebug() << "ResultsToSpectrum::addPath(" << xml << ")";
	QDomDocument doc( "bla" );
	doc.setContent( xml );
	fromXml( doc.documentElement() );
}

bool ResultsToSpectrum::restoreFromXML( QDomNode node ) { return node.nodeName() == "resultstospectrum"; }
bool ResultsToSpectrum::createFromXML( QDomNode node ) {
	if( thread() != QThread::currentThread() )
		qFatal( "ResultsToSpectrum::createFromXML() Not in my own thread!" );
	//qDebug() << "ResultsToSpectrum::createFromXML(" << &node << ")";
	if ( node.nodeName() == "resultstospectrum" ) {

		QDomNode elem = node.firstChildElement( "spectrum" );
		while ( ! elem.isNull() ) {

			NSCore::Thread_Template<ResultAction>* thread = new NSCore::Thread_Template<ResultAction>( this );
			thread->start();
			thread->waitForStarted();
			ResultAction* action = thread->item();
			connect( action, SIGNAL( stopMe() ), thread, SLOT( quit() ) );
			connect( this, SIGNAL( data( NSCore::AcqisitionDataSet ) ), action, SLOT( data( NSCore::AcqisitionDataSet ) ) );
			addChild( action );
			action->fromXml( elem );

			emit newSpectrum( action->_model );

			elem = elem.nextSiblingElement( "spectrum" );
		}

	}
	//qDebug() << "ResultsToSpectrum::createFromXML()  finished!";
	return false;
}

QStringList ResultsToSpectrum::spectra() const {
	QStringList ret = childlist<ResultAction>();
	qDebug() << "ResultsToSpectrum::spectra() returning: " << ret.join( "," );
	return ret;
}

AbstractSpectrumModel* ResultsToSpectrum::spectrum( QString name ) {
	qDebug() << "ResultsToSpectrum::spectrum(" << name << ")";
	ResultAction* action = child<ResultAction>( name );
	if ( action )
		return action->_model;
	return 0;
}

bool ResultsToSpectrum::dataArrived( NSCore::AcqisitionDataSet& d ) {
	//qDebug() << "ResultsToSpectrum::dataArrived(" << d << ") [defunct]";
	emit data( d );
	return true;
}

void ResultsToSpectrum::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	qDebug() << "ResultsToSpectrum::guiExecAction(" << action << "," << gui << ")";
	if ( action == "Add Spectrum" ) {
		QString name = gui->getString( "Spectrum name", "Please enter the name of the new spectrum", "" );
		if ( name.isNull() )
			return;

		int dimensions = gui->getInt( "Dimensions", "Please enter the number of (base) dimensions of the spectrum. Currently 1 and 2 are supported.", 1, 1, 2 );
		if ( dimensions > 2 )
			return;

		addPath( QString( "<resultstospectrum><spectrum name=\"%1\" dimensions=\"%2\"/></resultstospectrum>" ).arg( name ).arg( dimensions ) );
	}
}

QVariant ResultsToSpectrum::guiIdentifier( int role ) const {
	switch ( role ) {
		case Qt::DecorationRole:
			return QIcon( ":/icons/kde/cr48-apps-edu_mathematics.png" );
			break;
		default:
			return QTreeItem::guiIdentifier( role );
	}
}

